@@ -4,9 +4,10 @@ import asyncio as aio
 import dataclasses as dtc
 import functools as fnt
 import itertools as itt
+import collections as clc
 import typing as ty
-
-import util
+import time
+from aurcore import util
 
 
 class Event(util.AutoRepr):
@@ -23,44 +24,64 @@ class Event(util.AutoRepr):
       self.name = Event.hoist_name(self.name, router)
 
 
-@dtc.dataclass(frozen=True)
+# @dtc.dataclass(frozen=True)
 class EventWaiter:
    future: aio.Future
-   check: ty.Callable[[Event], ty.Coroutine[bool]]
+
+   def __init__(self, check: ty.Callable[[Event], ty.Awaitable[bool]], timeout: ty.Optional[float], max_matches: ty.Optional[int]):
+      self.check = check
+      self.timeout = timeout
+      self.start = time.perf_counter()
+      self.max_results = max_matches
+      self.queue = aio.Queue()
+      self.done = False
+
+   async def listener(self, event: Event):
+      if self.done:
+         return True
+      if self.timeout is not None and (time.perf_counter() - self.start) > self.timeout:
+         raise aio.TimeoutError()
+      if await self.check(event):
+         await self.queue.put(event)
+
+   async def producer(self):
+      try:
+         results = 0
+         while self.max_results is None or results < self.max_results:
+            yield await self.queue.get()
+            if self.max_results:
+               results += 1
+      except GeneratorExit:
+         self.done = True
+         raise GeneratorExit()
 
 
 class Eventful(util.AutoRepr):
-   f: ty.Callable
-   EventableFunc: ty.TypeAlias = ty.Callable[[Event], ty.Coroutine]
-   Eventable: ty.TypeAlias = ty.Union[EventableFunc, EventWaiter]
+   EventableFunc: ty.TypeAlias = ty.Callable[[Event], ty.Awaitable[ty.Optional[bool]]]
+   f: EventableFunc
 
-   def __init__(self, muxer: EventMuxer, eventable: Eventable):
+   def __init__(self, muxer: EventMuxer, eventable: EventableFunc):
       self.retain = True
       self.muxer = muxer
-      if isinstance(eventable, EventWaiter):
-         async def __waiter_wrapper(event: Event):
-            if eventable.future.cancelled():
-               self.retain = False
-            elif await eventable.check(event):
-               eventable.future.set_result(event)
-               self.retain = False
+      self.f = util.coroify(eventable)
+      self.f_orig = self.f
 
-         self.f = __waiter_wrapper
-      else:
-         self.f = util.coroify(eventable)
-
-   def __call__(self, event: Event) -> ty.Coroutine:
-      async def __retain_wrapper(ev: Event):
-         await self.f(ev)
-         return self.retain
-
-      return __retain_wrapper(event)
+   def __call__(self, event: Event) -> ty.Awaitable[ty.Optional[bool]]:
+      # Listeners return True to delete themselves, anything else (None) to
+      async def should_retain():
+         should_delete = await self.f(event)
+         if should_delete is True:
+            return False
+         if should_delete is None:
+            return True
+         raise RuntimeError(f"{self.f} returned something other than [True, None]")
+      return should_retain()
 
    @staticmethod
-   def decompose(func: ty.Callable[[...], ty.Coroutine]) -> EventableFunc:
+   def decompose(func: ty.Callable[[...], ty.Awaitable[None]]) -> ty.Callable[[...], ty.Awaitable[None]]:
       @fnt.wraps(func)
-      def __decompose_wrapper(event: Event):
-         return func(*event.args, **event.kwargs)
+      async def __decompose_wrapper(event: Event):
+         await func(*event.args, **event.kwargs)
 
       return __decompose_wrapper
 
@@ -91,8 +112,8 @@ class EventMuxer(util.AutoRepr):
 
 class EventRouterHost(util.AutoRepr):
    def __init__(self, name: ty.Optional[str] = "Unnamed"):
-      self.name = name
-      self.routers: ty.Dict[str, EventRouter] = {}
+      self.name = name.lower()
+      self.routers: ty.Dict[str, ty.List[EventRouter]] = clc.defaultdict(list)
 
    def __str__(self):
       return f"EventRouterHost {self.name} | Routers: {self.routers}"
@@ -100,28 +121,27 @@ class EventRouterHost(util.AutoRepr):
    def register(self, router: EventRouter) -> None:
       if router.name in self.routers:
          raise RuntimeError(f"[{self}] already has an event router named {router.name}")
-      self.routers[router.name] = router
+      self.routers[router.name].append(router)
 
    def deregister(self, router: EventRouter) -> None:
-      if router.name in self.routers:
-         del self.routers[router.name]
+      if router in self.routers:
+         self.routers[router.name].remove(router)
       else:
          raise RuntimeError(f"[{self}] attempted to deregister an unregistered router {router}")
 
    # noinspection PyProtectedMember
    async def submit(self, event: Event):
-      await aio.gather(*[router._dispatch(event) for router in self.routers.values()])
-
+      await aio.gather(*[router._dispatch(event) for router_group in self.routers.values() for router in router_group])
 
 
 class EventRouter(util.AutoRepr):
    def __init__(self, name: str, host: EventRouterHost):
-      self.name = name
+      self.name = name.lower()
       self.host = host
       self.host.register(self)
       self.muxers: ty.Dict[str, EventMuxer] = {}
 
-   def _register_listener(self, event_name: str, listener: Eventful.Eventable):
+   def _register_listener(self, event_name: str, listener: Eventful.EventableFunc):
       event_name = Event.hoist_name(event_name.lower(), self)
 
       if event_name not in self.muxers:
@@ -135,19 +155,19 @@ class EventRouter(util.AutoRepr):
       event_name = Event.hoist_name(event_name.lower(), self)
 
       def listen_deco(func: Eventful.EventableFunc):
+         func_: Eventful.EventableFunc = util.coroify(func)
          if decompose:
-            self._register_listener(event_name, Eventful.decompose(func))
+            self._register_listener(event_name, Eventful.decompose(func_))
          else:
-            self._register_listener(event_name, func)
+            self._register_listener(event_name, func_)
          return func
 
       return listen_deco
 
-   def wait_for(self, event_name: str, check: ty.Callable[[Event], bool], timeout: float) -> aio.Future:
-      fut = aio.Future()
-      ev_waiter: EventWaiter = EventWaiter(future=fut, check=util.coroify(check))
-      self._register_listener(event_name=event_name, listener=ev_waiter)
-      return aio.wait_for(fut, timeout=timeout)
+   def wait_for(self, event_name: str, check: ty.Callable[[Event], bool], timeout: float = None, max_matches=1) -> util.AwaitableAiter:
+      ev_waiter: EventWaiter = EventWaiter(check=util.coroify(check), timeout=timeout, max_matches=max_matches)
+      self._register_listener(event_name=event_name, listener=ev_waiter.listener)
+      return util.AwaitableAiter(ev_waiter.producer())
 
    async def submit(self, event: Event) -> None:
       event.hoist(self)
@@ -157,4 +177,7 @@ class EventRouter(util.AutoRepr):
       await aio.gather(*[muxer.fire(event) for listen_name, muxer in self.muxers.items() if event.name.startswith(listen_name)])
 
    def detach(self):
-      self.host.deregister(self)
\ No newline at end of file
+      self.host.deregister(self)
+
+   def __str__(self):
+      return f"[Muxers: <{self.muxers}>]"
\ No newline at end of file
